// MyOcean implementation
// Copyright (C) 01.09.97 Marco Dettori and Oliver Obst
//
// $Id: myocean.C,v 0.2 1998/03/16 10:31:42 oliver Exp $

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

#include <iostream.h>

#include "myocean.h"
#include "seaproto.h"

// --- constructor ------------------------------------------------------------

MyOcean :: MyOcean( char sql, char ships, char *lengths, char stlen,
                    QWidget *parent, const char *name )
        : Ocean( sql, ships, lengths, stlen, parent, name )
{
        first   = ship  = -1;
        len     = 0;
        num     = ships;
        length  = lengths;
        connect( this, SIGNAL( shipReady( int ) ), SLOT( setNext() ) );
        emit shipReady( ship );
}

// --- public slot -----------------------------------------------------------
// MyOcean :: reset()
//
// set all ships again, assuming number and lengths of the ships are known

void MyOcean :: reset()
{
        if ( num > 0 ) {
                first   = ship  = -1;

                for (int i=0; i < size; i++ ) {
                        buttons->at(i)->setState( TriState::EMPTY );
                        buttons->at(i)->setBackgroundColor( blue );
                        buttons->at(i)->visit( FALSE );
                }

                st->resetStatus();
                emit shipReady( ship );                
        }
}

// --- public slot -----------------------------------------------------------
//
// MyOcean :: clearWater
//
// reset all CIRCLE-Buttons to EMPTY; leave CROSS-Buttons as they are

void MyOcean :: clearWater()
{
        for (int i=0; i < size; i++)
                if ( buttons->at(i)->getState() == TriState::CIRCLE )
                        buttons->at(i)->setState( TriState::EMPTY );
}


// --- protected -------------------------------------------------------------
// MyOcean :: checkShip
//
// check if a ship (length l) can be set to id1, in direction over id2
// returns TRUE or FALSE
// example: length 4
//          0   X   X   X   X   0
//             id1 id2

bool MyOcean :: checkShip( short id1, short id2, char l )
{
        bool    ok;
        short   diff = id2 - id1;
        short   line = int ( id1 / edge ) * edge;

        ok = ( ( ( diff == -1 ) && ( id1 - l >= line ) ) ||
               ( ( diff ==  1 ) && ( id1 + l < line + edge ) ) ||
               ( ( diff == -edge ) && ( id1 - l * edge >= 0 ) ) ||
               ( ( diff == edge ) && ( id1 + l * edge < size ) ) );

        short i = 1;
        while ( ok && i<= l ) {
                ok = (buttons->at(id1 + diff * i)->getState()
                      == TriState::EMPTY);
                i++;
        }
        return ok;
}

// --- protected -------------------------------------------------------------
// MyOcean :: setShip
//

void MyOcean :: setShip( short id1, short id2, char l )
{
        short   diff = id2 - id1;

        for (int i=1; i <= l; i++)
                buttons->at(id1 + diff * i)->setState(TriState::CROSS);
}

// --- protected --------------------------------------------------------------
// MyOcean::survivors
//

int MyOcean::survivors( int id )
{
        int     line = int ( id / edge ) * edge;
        int     rest = 0;
        int     i;

        i = 1;
        while (id - i >= line &&
               buttons->at(id - i)->getState() == TriState::CROSS) {
                if (! buttons->at(id - i)->visited())
                        rest++;
                i++;
        }

        i = 1;
        while ( id + i < line + edge &&
                buttons->at(id + i)->getState() == TriState::CROSS) {
                if ( !buttons->at(id + i)->visited() )
                        rest++;
                i++;
        }

        i = edge;
        while (id - i >= 0 &&
               buttons->at(id - i)->getState() == TriState::CROSS) {
                if (! buttons->at(id - i)->visited())
                        rest++;
                i += edge;
        }

        i = edge;
        while (id + i < size &&
               buttons->at(id + i)->getState() == TriState::CROSS) {
                if (! buttons->at(id + i)->visited())
                        rest++;
                i += edge;
        }

        return rest;
}


// --- public slot ------------------------------------------------------------
// MyOcean::peek( int id )
//
// peek into a square

void MyOcean :: peek( int id )
{
        TriState::State s = buttons->at(id)->getState();

        buttons->at(id)->setBackgroundColor( red );
        buttons->at(id)->visit();

        if ( s == TriState::CROSS ) {
                if ( survivors( id ) == 0 ) {
                        if ( --ship == 0 )
                                emit reporter( id, DEFEAT );
                        else
                                emit reporter( id, SINK );
                }
                else
                        emit reporter( id, HIT );
        } else {
                emit reporter( id, MISSED );
        }
}


// --- protected slot ---------------------------------------------------------
// MyOcean :: handleClick
//
// what to do if button with ID id is pressed 

void MyOcean :: handleClick()
{
        int id = buttons->findRef((TriState *) sender());        
        TriState::State s = buttons->at(id)->getState();

        if ( len > 0 ) {
                if ( first == -1) {  // the first box of the ship has to be set
                        if ( s == TriState::EMPTY ) {
                                buttons->at(id)->setState( TriState::CROSS );
                                first = id;
                                len--;
                        }
                } else { // setting a box (but not the first one)
                        if ( s == TriState::CROSS && first == id ) {
                                buttons->at(id)->setState(TriState::EMPTY);
                                first = -1;
                                len++;
                        } else {
                                if ( checkShip( first, id, len ) ) {
                                        setShip( first, id, len );
                                        len = 0;
                                } // if ( checkShip )
                        }  // else ( s == TriState::CROSS && first == id )
                } // else ( first == -1 )

                if ( len == 0 ) {
                        if ( first != -1 )
                                surround( id );
                        first = -1;
                        emit shipReady( ship );
                }
                
        }  // if ( len > 0 )
}

// --- protected slot ---------------------------------------------------------
// MyOcean :: setNext
//
// set length of the next ship to put, if there is any

void MyOcean :: setNext()
{
        st->setStatus( ship );
        
        if ( ++ship < num) 
                len = length[ ship ];
        else {
                len = 0;
                emit ready();
        }
}
